# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import os
import subprocess
from pathlib import Path
from typing import List, Literal, Optional, Union

import torch
from omegaconf import OmegaConf

from netspresso_trainer.trainer_common import train_common
from netspresso_trainer.utils.engine_utils import (
    LOG_LEVEL,
    get_gpus_from_parser_and_config,
    parse_args_netspresso,
    parse_gpu_ids,
    set_arguments,
    validate_train_config,
)


def run_distributed_training_script(gpu_ids, data, augmentation, model, training, logging, environment, log_level,
                                    task, model_name, is_graphmodule_training, logging_dir):

    command = [
        "--data", data,
        "--augmentation", augmentation,
        "--model", model,
        "--training", training,
        "--logging", logging,
        "--environment", environment,
        "--log-level", log_level,
        "--task", task,
        "--model-name", model_name,
        "--is-graphmodule-training", is_graphmodule_training,
        "--logging-dir", logging_dir,
    ]

    # Distributed training script
    command = [
        'python', '-m', 'torch.distributed.launch',
        f'--nproc_per_node={len(gpu_ids)}',  # GPU #
        f"{Path(__file__).absolute().parent / 'trainer_main.py'}", *map(str, command)
    ]

    # Run subprocess
    process = subprocess.Popen(command)

    try:
        process.wait()
    except KeyboardInterrupt:
        print("Interrupted. Terminating the training process...")
        process.terminate()
        process.wait()


def train_with_yaml_impl(gpus: Optional[Union[List, int]], data: Union[Path, str], augmentation: Union[Path, str],
                         model: Union[Path, str], training: Union[Path, str],
                         logging: Union[Path, str], environment: Union[Path, str], log_level: str = LOG_LEVEL):
    conf_environment = OmegaConf.load(environment).environment
    gpus = get_gpus_from_parser_and_config(gpus, conf_environment)
    assert isinstance(gpus, (list, int))

    gpu_ids_str = ','.join(map(str, gpus)) if isinstance(gpus, list) else str(gpus)
    os.environ['CUDA_VISIBLE_DEVICES'] = gpu_ids_str
    torch.cuda.empty_cache()  # Reinitialize CUDA to apply the change

    conf = set_arguments(data=data,
                         augmentation=augmentation,
                         model=model,
                         training=training,
                         logging=logging,
                         environment=environment)
    config_summary = validate_train_config(conf)

    try:
        if isinstance(gpus, int):
            train_common(
                conf,
                task=config_summary.task,
                model_name=config_summary.model_name,
                is_graphmodule_training=config_summary.is_graphmodule_training,
                logging_dir=config_summary.logging_dir,
                log_level=log_level
            )
        else:
            run_distributed_training_script(
                gpus, data, augmentation, model, training, logging, environment, log_level,
                config_summary.task, config_summary.model_name, config_summary.is_graphmodule_training, config_summary.logging_dir
            )
        return config_summary.logging_dir
    except Exception as e:
        raise e


def train_with_yaml(
    data: Union[Path, str],
    augmentation: Union[Path, str],
    model: Union[Path, str], training: Union[Path, str],
    logging: Union[Path, str], environment: Union[Path, str],
    gpus: Optional[str] = None, log_level: str = LOG_LEVEL
):

    gpus: Union[List, int] = parse_gpu_ids(gpus)

    logging_dir: Path = train_with_yaml_impl(
        gpus=gpus,
        data=data,
        augmentation=augmentation,
        model=model,
        training=training,
        logging=logging,
        environment=environment,
        log_level=log_level
    )

    return logging_dir


def train_cli() -> None:
    args_parsed = parse_args_netspresso(with_gpus=True, isTrain=True)

    logging_dir: Path = train_with_yaml_impl(
        gpus=args_parsed.gpus,
        data=args_parsed.data,
        augmentation=args_parsed.augmentation,
        model=args_parsed.model,
        training=args_parsed.training,
        logging=args_parsed.logging,
        environment=args_parsed.environment,
        log_level=args_parsed.log_level
    )

    return logging_dir


def train_cli_without_additional_gpu_check() -> None:
    args_parsed = parse_args_netspresso(with_gpus=False, isTrain=True)

    conf = set_arguments(
        data=args_parsed.data,
        augmentation=args_parsed.augmentation,
        model=args_parsed.model,
        training=args_parsed.training,
        logging=args_parsed.logging,
        environment=args_parsed.environment
    )

    train_common(
        conf,
        task=args_parsed.task,
        model_name=args_parsed.model_name,
        is_graphmodule_training=args_parsed.is_graphmodule_training,
        logging_dir=args_parsed.logging_dir,
        log_level=args_parsed.log_level
    )

if __name__ == "__main__":

    # Execute by `run_distributed_training_script`
    train_cli_without_additional_gpu_check()



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import math

import torch
import torch.distributed as dist
from torch.utils.data import Sampler


class DistributedEvalSampler(Sampler):
    r"""
    DistributedEvalSampler is different from DistributedSampler.
    It does NOT add extra samples to make it evenly divisible.
    DistributedEvalSampler should NOT be used for training. The distributed processes could hang forever.
    See this issue for details: https://github.com/pytorch/pytorch/issues/22584
    shuffle is disabled by default

    DistributedEvalSampler is for evaluation purpose where synchronization does not happen every epoch.
    Synchronization should be done outside the dataloader loop.

    Sampler that restricts data loading to a subset of the dataset.

    It is especially useful in conjunction with
    :class:`torch.nn.parallel.DistributedDataParallel`. In such a case, each
    process can pass a :class`~torch.utils.data.DistributedSampler` instance as a
    :class:`~torch.utils.data.DataLoader` sampler, and load a subset of the
    original dataset that is exclusive to it.

    .. note::
        Dataset is assumed to be of constant size.

    Arguments:
        dataset: Dataset used for sampling.
        num_replicas (int, optional): Number of processes participating in
            distributed training. By default, :attr:`rank` is retrieved from the
            current distributed group.
        rank (int, optional): Rank of the current process within :attr:`num_replicas`.
            By default, :attr:`rank` is retrieved from the current distributed
            group.
        shuffle (bool, optional): If ``True`` (default), sampler will shuffle the
            indices.
        seed (int, optional): random seed used to shuffle the sampler if
            :attr:`shuffle=True`. This number should be identical across all
            processes in the distributed group. Default: ``0``.

    .. warning::
        In distributed mode, calling the :meth`set_epoch(epoch) <set_epoch>` method at
        the beginning of each epoch **before** creating the :class:`DataLoader` iterator
        is necessary to make shuffling work properly across multiple epochs. Otherwise,
        the same ordering will be always used.

    Example::

        >>> sampler = DistributedSampler(dataset) if is_distributed else None
        >>> loader = DataLoader(dataset, shuffle=(sampler is None),
        ...                     sampler=sampler)
        >>> for epoch in range(start_epoch, n_epochs):
        ...     if is_distributed:
        ...         sampler.set_epoch(epoch)
        ...     train(loader)
    """

    def __init__(self, dataset, num_replicas=None, rank=None, shuffle=False, seed=0):
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = dist.get_rank()
        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0
        # self.num_samples = int(math.ceil(len(self.dataset) * 1.0 / self.num_replicas))
        # self.total_size = self.num_samples * self.num_replicas
        self.num_samples = math.ceil(len(self.dataset) / self.num_replicas)             # true value without extra samples
        self.total_size = self.num_samples * self.num_replicas

        self.shuffle = shuffle
        self.seed = seed

    def __iter__(self):
        if self.shuffle:
            # deterministically shuffle based on epoch and seed
            g = torch.Generator()
            g.manual_seed(self.seed + self.epoch)
            indices = torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = list(range(len(self.dataset)))


        # # add extra samples to make it evenly divisible
        # indices += indices[:(self.total_size - len(indices))]
        # assert len(indices) == self.total_size
        indices += [-1 for _ in range(self.total_size - len(indices))]

        # subsample
        indices = indices[self.rank:self.total_size:self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self):
        return self.num_samples

    def set_epoch(self, epoch):
        r"""
        Sets the epoch for this sampler. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.

        Arguments:
            epoch (int): _epoch number.
        """
        self.epoch = epoch



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Literal, Optional

import torch

from .base import BaseTaskProcessor


class DetectionProcessor(BaseTaskProcessor):
    def __init__(self, conf, postprocessor, devices, **kwargs):
        super(DetectionProcessor, self).__init__(conf, postprocessor, devices, **kwargs)
        self.num_classes = kwargs['num_classes']

    def train_step(self, train_model, batch, optimizer, loss_factory, metric_factory):
        train_model.train()
        images, labels, bboxes = batch['pixel_values'], batch['label'], batch['bbox']
        images = images.to(self.devices)
        targets = [{"boxes": box.to(self.devices), "labels": label.to(self.devices),}
                   for box, label in zip(bboxes, labels)]

        targets = {'gt': targets,
                   'img_size': images.size(-1),
                   'num_classes': self.num_classes,}

        optimizer.zero_grad()

        with torch.cuda.amp.autocast(enabled=self.mixed_precision):
            out = train_model(images, targets=targets)
            loss_factory.calc(out, targets, phase='train')

        loss_factory.backward(self.grad_scaler)
        if self.max_norm:
            # Unscales the gradients of optimizer's assigned parameters in-place
            self.grad_scaler.unscale_(optimizer)
            torch.nn.utils.clip_grad_norm_(train_model.parameters(), self.max_norm)
        # optimizer's gradients are already unscaled, so scaler.step doesn't unscale them,
        self.grad_scaler.step(optimizer)
        self.grad_scaler.update()

        pred = self.postprocessor(out, original_shape=images[0].shape)

        if self.conf.distributed:
            gathered_bboxes = [None for _ in range(torch.distributed.get_world_size())]
            gathered_labels = [None for _ in range(torch.distributed.get_world_size())]
            gathered_pred = [None for _ in range(torch.distributed.get_world_size())]
            torch.distributed.gather_object(bboxes, gathered_bboxes if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.gather_object(labels, gathered_labels if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.gather_object(pred, gathered_pred if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.barrier()
            if torch.distributed.get_rank() == 0:
                gathered_bboxes = sum(gathered_bboxes, [])
                gathered_labels = sum(gathered_labels, [])
                gathered_pred = sum(gathered_pred, [])
                bboxes = gathered_bboxes
                labels = gathered_labels
                pred = gathered_pred

        if self.single_gpu_or_rank_zero:
            logs = {
                'target': [(bbox.detach().cpu().numpy(), label.detach().cpu().numpy())
                        for bbox, label in zip(bboxes, labels)],
                'pred': pred
            }
            return dict(logs.items())

    def valid_step(self, eval_model, batch, loss_factory, metric_factory):
        eval_model.eval()
        indices, images, labels, bboxes = batch['indices'], batch['pixel_values'], batch['label'], batch['bbox']
        images = images.to(self.devices)
        targets = [{"boxes": box.to(self.devices), "labels": label.to(self.devices)}
                   for box, label in zip(bboxes, labels)]

        targets = {'gt': targets,
                   'img_size': images.size(-1),
                   'num_classes': self.num_classes,}

        out = eval_model(images)
        loss_factory.calc(out, targets, phase='valid')

        pred = self.postprocessor(out, original_shape=images[0].shape)

        indices = indices.numpy()
        if self.conf.distributed:
            # Remove dummy samples, they only come in distributed environment
            images = images[indices != -1]
            filtered_bboxes = []
            filtered_labels = []
            filtered_pred = []
            for idx, bool_idx in enumerate(indices != -1):
                if bool_idx:
                    filtered_bboxes.append(bboxes[idx])
                    filtered_labels.append(labels[idx])
                    filtered_pred.append(pred[idx])
            bboxes = filtered_bboxes
            labels = filtered_labels
            pred = filtered_pred

            # Gather phase
            gathered_bboxes = [None for _ in range(torch.distributed.get_world_size())]
            gathered_labels = [None for _ in range(torch.distributed.get_world_size())]
            gathered_pred = [None for _ in range(torch.distributed.get_world_size())]
            torch.distributed.gather_object(bboxes, gathered_bboxes if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.gather_object(labels, gathered_labels if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.gather_object(pred, gathered_pred if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.barrier()
            if torch.distributed.get_rank() == 0:
                gathered_bboxes = sum(gathered_bboxes, [])
                gathered_labels = sum(gathered_labels, [])
                gathered_pred = sum(gathered_pred, [])
                bboxes = gathered_bboxes
                labels = gathered_labels
                pred = gathered_pred

        if self.single_gpu_or_rank_zero:
            logs = {
                'images': images.detach().cpu().numpy(),
                'target': [(bbox.detach().cpu().numpy(), label.detach().cpu().numpy())
                        for bbox, label in zip(bboxes, labels)],
                'pred': pred
            }
            return dict(logs.items())

    def test_step(self, test_model, batch):
        test_model.eval()
        indices, images = batch['indices'], batch['pixel_values']
        images = images.to(self.devices)

        out = test_model(images)

        pred = self.postprocessor(out, original_shape=images[0].shape)

        indices = indices.numpy()
        if self.conf.distributed:
            # Remove dummy samples, they only come in distributed environment
            filtered_pred = []
            for idx, bool_idx in enumerate(indices != -1):
                if bool_idx:
                    filtered_pred.append(pred[idx])
            pred = filtered_pred

            # Gather phase
            gathered_pred = [None for _ in range(torch.distributed.get_world_size())]
            torch.distributed.gather_object(pred, gathered_pred if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.barrier()
            if torch.distributed.get_rank() == 0:
                gathered_pred = sum(gathered_pred, [])
                pred = gathered_pred

        if self.single_gpu_or_rank_zero:
            results = {'images': images.detach().cpu().numpy(), 'pred': pred}
            return results

    def get_metric_with_all_outputs(self, outputs, phase: Literal['train', 'valid'], metric_factory):
        if self.single_gpu_or_rank_zero:
            pred = []
            targets = []
            for output_batch in outputs:
                if len(output_batch['target']) == 0:
                    continue

                for detection, class_idx in output_batch['target']:
                    target_on_image = {}
                    target_on_image['boxes'] = detection
                    target_on_image['labels'] = class_idx
                    targets.append(target_on_image)

                for detection, class_idx in output_batch['pred']:
                    pred_on_image = {}
                    pred_on_image['post_boxes'] = detection[..., :4]
                    pred_on_image['post_scores'] = detection[..., -1]
                    pred_on_image['post_labels'] = class_idx
                    pred.append(pred_on_image)
            metric_factory.update(pred, target=targets, phase=phase)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import copy
import json
from ctypes import c_int
from dataclasses import asdict
from pathlib import Path
from statistics import mean
from typing import Dict, List, Literal, Optional, final

import torch
import torch.nn as nn
from loguru import logger
from omegaconf import DictConfig
from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader
from tqdm import tqdm

from ..loggers.base import TrainingLogger
from ..losses.builder import LossFactory
from ..metrics.builder import MetricFactory
from ..utils.checkpoint import load_checkpoint, save_checkpoint
from ..utils.fx import save_graphmodule
from ..utils.logger import yaml_for_logging
from ..utils.model_ema import ModelEMA
from ..utils.onnx import save_onnx
from ..utils.record import Timer, TrainingSummary
from ..utils.stats import get_params_and_flops
from .base import BasePipeline
from .task_processors.base import BaseTaskProcessor

NUM_SAMPLES = 16


class TrainingPipeline(BasePipeline):
    def __init__(
        self,
        conf: DictConfig,
        task: str,
        task_processor: BaseTaskProcessor,
        model_name: str,
        model: nn.Module,
        logger: Optional[TrainingLogger],
        timer: Timer,
        optimizer: Optimizer,
        scheduler: _LRScheduler,
        loss_factory: LossFactory,
        metric_factory: MetricFactory,
        train_dataloader: DataLoader,
        eval_dataloader: DataLoader,
        single_gpu_or_rank_zero: bool,
        is_graphmodule_training: bool,
        model_ema: Optional[ModelEMA],
        start_epoch: int,
        cur_epoch: c_int,
        profile: bool,
    ):
        super(TrainingPipeline, self).__init__(conf, task, task_processor, model_name, model, logger, timer)
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.loss_factory = loss_factory
        self.metric_factory = metric_factory
        self.train_dataloader = train_dataloader
        self.eval_dataloader = eval_dataloader
        self.single_gpu_or_rank_zero = single_gpu_or_rank_zero
        self.is_graphmodule_training = is_graphmodule_training
        self.model_ema = model_ema
        self.start_epoch = start_epoch
        self.cur_epoch = cur_epoch
        self.profile = profile  # TODO: provide torch_tb_profiler for training

        self.training_history: Dict[int, Dict[
            Literal['train_losses', 'valid_losses', 'train_metrics', 'valid_metrics'], Dict[str, float]
        ]] = {}

        # TODO: These will be removed
        self.save_optimizer_state = True

    @final
    def _is_ready(self):
        assert self.model is not None, "`self.model` is not defined!"
        assert self.optimizer is not None, "`self.optimizer` is not defined!"
        """Append here if you need more assertion checks!"""
        assert self.conf.logging.save_checkpoint_epoch % self.conf.logging.validation_epoch == 0, \
            "`save_checkpoint_epoch` should be the multiplier of `validation_epoch`."
        return True

    def epoch_with_valid_logging(self, epoch: int):
        validation_freq = self.conf.logging.validation_epoch
        last_epoch = epoch == self.conf.training.epochs
        return (epoch % validation_freq == 1 % validation_freq) or last_epoch

    def epoch_with_checkpoint_saving(self, epoch: int):
        checkpoint_freq = self.conf.logging.save_checkpoint_epoch
        last_epoch = epoch == self.conf.training.epochs
        return (epoch % checkpoint_freq == 1 % checkpoint_freq) or last_epoch

    @property
    def learning_rate(self):
        return mean([param_group['lr'] for param_group in self.optimizer.param_groups])

    @property
    def train_loss(self):
        return self.loss_factory.result('train').get('total').avg

    @property
    def valid_loss(self):
        return self.loss_factory.result('valid').get('total').avg

    def train(self):
        if self.single_gpu_or_rank_zero:
            logger.debug(f"Training configuration:\n{yaml_for_logging(self.conf)}")
            logger.info("-" * 40)

        self.timer.start_record(name='train_all')
        self._is_ready()

        num_epoch = -1
        try:
            for num_epoch in range(self.start_epoch, self.conf.training.epochs + 1):
                self.timer.start_record(name=f'train_epoch_{num_epoch}')
                self.loss_factory.reset_values()
                self.metric_factory.reset_values()
                self.cur_epoch.value = num_epoch

                self.train_one_epoch(epoch=num_epoch)

                with_valid_logging = self.epoch_with_valid_logging(num_epoch)
                with_checkpoint_saving = self.epoch_with_checkpoint_saving(num_epoch)
                # FIXME: multi-gpu sample counting & validation
                valid_samples = self.validate() if with_valid_logging else None

                self.timer.end_record(name=f'train_epoch_{num_epoch}')
                time_for_epoch = self.timer.get(name=f'train_epoch_{num_epoch}', as_pop=False)

                if self.single_gpu_or_rank_zero:
                    self.log_end_epoch(epoch=num_epoch,
                                       time_for_epoch=time_for_epoch,
                                       valid_samples=valid_samples,
                                       valid_logging=with_valid_logging)
                    self.save_summary()
                    if with_checkpoint_saving:
                        assert with_valid_logging
                        self.save_checkpoint(epoch=num_epoch)
                    logger.info("-" * 40)

                self.scheduler.step()  # call after reporting the current `learning_rate`

            self.timer.end_record(name='train_all')
            total_train_time = self.timer.get(name='train_all', as_pop=False)
            logger.info(f"Total time: {total_train_time:.2f} s")

            if self.single_gpu_or_rank_zero:
                self.logger.log_end_of_traning(final_metrics={'time_for_last_epoch': time_for_epoch})
                self.save_best()
                self.save_summary(end_training=True, status="success")
        except KeyboardInterrupt as e:
            # TODO: add independent procedure for KeyboardInterupt
            logger.error("Keyboard interrupt detected! Try saving the current checkpoint...")
            if self.single_gpu_or_rank_zero:
                self.save_checkpoint(epoch=num_epoch)
                self.save_best()
                self.save_summary(status="stop")
            raise e
        except Exception as e:
            logger.error("Error occurred! Try saving the current checkpoint...")
            if self.single_gpu_or_rank_zero:
                self.save_checkpoint(epoch=num_epoch)
                self.save_best()
                self.save_summary(status="error", error_stats=str(e))
            logger.error(str(e))
            raise e

    def train_one_epoch(self, epoch):
        outputs = []
        for _idx, batch in enumerate(tqdm(self.train_dataloader, leave=False)):
            out = self.task_processor.train_step(self.model, batch, self.optimizer, self.loss_factory, self.metric_factory)
            if self.model_ema:
                self.model_ema.update(model=self.model.module if hasattr(self.model, 'module') else self.model)
            outputs.append(out)
        self.task_processor.get_metric_with_all_outputs(outputs, phase='train', metric_factory=self.metric_factory)

    @torch.no_grad()
    def validate(self, num_samples=NUM_SAMPLES):
        num_returning_samples = 0
        returning_samples = []
        outputs = []
        eval_model = self.model_ema.ema_model if self.model_ema else self.model
        for _idx, batch in enumerate(tqdm(self.eval_dataloader, leave=False)):
            out = self.task_processor.valid_step(eval_model, batch, self.loss_factory, self.metric_factory)
            if out is not None:
                outputs.append(out)
                if num_returning_samples < num_samples:
                    returning_samples.append(out)
                    num_returning_samples += len(out['pred'])
        self.task_processor.get_metric_with_all_outputs(outputs, phase='valid', metric_factory=self.metric_factory)
        return returning_samples

    def log_end_epoch(
        self,
        epoch: int,
        time_for_epoch: float,
        valid_samples: Optional[List] = None,
        valid_logging: bool = False,
    ):
        train_losses = self.loss_factory.result('train')
        train_metrics = self.metric_factory.result('train')
        self.log_results(prefix='training', epoch=epoch, losses=train_losses, metrics=train_metrics,
                         learning_rate=self.learning_rate, elapsed_time=time_for_epoch)

        if valid_logging:
            valid_losses = self.loss_factory.result('valid') if valid_logging else None
            valid_metrics = self.metric_factory.result('valid') if valid_logging else None
            self.log_results(prefix='validation', epoch=epoch, samples=valid_samples, losses=valid_losses, metrics=valid_metrics)

        summary_record = {'train_losses': train_losses, 'train_metrics': train_metrics}
        if valid_logging:
            summary_record.update({'valid_losses': valid_losses, 'valid_metrics': valid_metrics})
        self.training_history.update({epoch: summary_record})

    def save_checkpoint(self, epoch: int):
        if self.model_ema:
            model = self.model_ema.ema_model
        else:
            model = self.model.module if hasattr(self.model, 'module') else self.model
        if hasattr(model, 'deploy'):
            model.deploy()
        save_dtype = model.save_dtype

        if save_dtype == torch.float16:
            model = copy.deepcopy(model).type(save_dtype)
        logging_dir = self.logger.result_dir
        model_path = Path(logging_dir) / f"{self.task}_{self.model_name}_epoch_{epoch}.ext"
        optimizer_path = Path(logging_dir) / f"{self.task}_{self.model_name}_epoch_{epoch}_optimzer.pth"

        if self.save_optimizer_state:
            optimizer = self.optimizer.module if hasattr(self.optimizer, 'module') else self.optimizer
            save_dict = {'optimizer': optimizer.state_dict(), 'last_epoch': epoch}
            torch.save(save_dict, optimizer_path)
            logger.debug(f"Optimizer state saved at {str(optimizer_path)}")

        if self.is_graphmodule_training:
            # Just save graphmodule checkpoint
            torch.save(model, model_path.with_suffix(".pt"))
            logger.debug(f"PyTorch FX model saved at {str(model_path.with_suffix('.pt'))}")
            return
        pytorch_model_state_dict_path = model_path.with_suffix(".safetensors")
        save_checkpoint(model.state_dict(), pytorch_model_state_dict_path)
        logger.debug(f"PyTorch model saved at {str(pytorch_model_state_dict_path)}")

    def save_best(self):
        valid_losses = {epoch: record['valid_losses'].get('total') for epoch, record in self.training_history.items()
                if 'valid_losses' in record}
        if not valid_losses:
            return # No validation loss recorded
        best_epoch = min(valid_losses, key=valid_losses.get)

        logging_dir = self.logger.result_dir

        best_checkpoint_path = Path(logging_dir) / f"{self.task}_{self.model_name}_epoch_{best_epoch}.ext"
        best_model_save_path = Path(logging_dir) / f"{self.task}_{self.model_name}_best.ext"

        model = self.model.module if hasattr(self.model, 'module') else self.model
        save_dtype = model.save_dtype
        best_model_to_save = copy.deepcopy(model)
        if hasattr(best_model_to_save, 'deploy'):
            best_model_to_save.deploy()

        if self.is_graphmodule_training:
            best_model_to_save.load_state_dict(load_checkpoint(best_checkpoint_path.with_suffix('.pt')).state_dict())
            save_onnx(best_model_to_save, best_model_save_path.with_suffix(".onnx"), sample_input=self.sample_input.type(save_dtype), opset_version=self.conf.logging.onnx_export_opset)
            logger.info(f"ONNX model converting and saved at {str(best_model_save_path.with_suffix('.onnx'))}")
            torch.save(best_model_to_save, best_model_save_path.with_suffix(".pt"))
            logger.info(f"Best model saved at {str(best_model_save_path.with_suffix('.pt'))}")
            return

        best_model_to_save.load_state_dict(load_checkpoint(best_checkpoint_path.with_suffix('.safetensors')))
        pytorch_best_model_state_dict_path = best_model_save_path.with_suffix(".safetensors")
        save_checkpoint(best_model_to_save.state_dict(), pytorch_best_model_state_dict_path)
        logger.info(f"Best model saved at {str(pytorch_best_model_state_dict_path)}")

        try:
            save_onnx(best_model_to_save, best_model_save_path.with_suffix(".onnx"), sample_input=self.sample_input.type(save_dtype), opset_version=self.conf.logging.onnx_export_opset)
            logger.info(f"ONNX model converting and saved at {str(best_model_save_path.with_suffix('.onnx'))}")

            save_graphmodule(best_model_to_save, (best_model_save_path.parent / f"{best_model_save_path.stem}_fx").with_suffix(".pt"))
            logger.info(f"PyTorch FX model tracing and saved at {str(best_model_save_path.with_suffix('.pt'))}")
        except Exception as e:
            logger.error(e)
            pass

    def save_summary(self, end_training=False, status="", error_stats=""):
        training_summary = TrainingSummary(
            total_epoch=self.conf.training.epochs,
            train_losses={epoch: record['train_losses'].get('total') for epoch, record in self.training_history.items()},
            valid_losses={epoch: record['valid_losses'].get('total') for epoch, record in self.training_history.items()
                          if 'valid_losses' in record},
            train_metrics={epoch: record['train_metrics'] for epoch, record in self.training_history.items()},
            valid_metrics={epoch: record['valid_metrics'] for epoch, record in self.training_history.items()
                           if 'valid_metrics' in record},
            metrics_list=self.metric_factory.metric_names,
            primary_metric=self.metric_factory.primary_metric,
        )
        if end_training:
            total_train_time = self.timer.get(name='train_all', as_pop=True)
            flops, params = get_params_and_flops(self.model, self.sample_input.float())
            logger.info(f"[Model stats] | Sample input: {tuple(self.sample_input.shape)} | Params: {(params/1e6):.2f}M | FLOPs: {(flops/1e9):.2f}G")
            training_summary.total_train_time = total_train_time
            training_summary.flops = flops
            training_summary.params = params

        training_summary.status = status
        training_summary.error_stats = error_stats

        logging_dir = self.logger.result_dir
        summary_path = Path(logging_dir) / "training_summary.json"

        with open(summary_path, 'w') as f:
            json.dump(asdict(training_summary), f, indent=4)
        logger.info(f"Model training summary saved at {str(summary_path)}")

    def profile_one_epoch(self):
        PROFILE_WAIT = 1
        PROFILE_WARMUP = 1
        PROFILE_ACTIVE = 10
        PROFILE_REPEAT = 1
        #_ = torch.ones(1).to(self.devices)
        with torch.profiler.profile(
            schedule=torch.profiler.schedule(wait=PROFILE_WAIT,
                                             warmup=PROFILE_WARMUP,
                                             active=PROFILE_ACTIVE,
                                             repeat=PROFILE_REPEAT),
            on_trace_ready=torch.profiler.tensorboard_trace_handler('./log/test'),
            record_shapes=True,
            profile_memory=True,
            with_flops=True,
            with_modules=True
        ) as prof:
            for idx, batch in enumerate(self.train_dataloader):
                if idx >= (PROFILE_WAIT + PROFILE_WARMUP + PROFILE_ACTIVE) * PROFILE_REPEAT:
                    break
                self.task_processor.train_step(batch)
                prof.step()



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import torch
from torch.nn import functional as F


def classification_mix_collate_fn(original_batch, mix_transforms):
    indices = []
    images = []
    target = []
    for data_sample in original_batch:
        indices.append(data_sample[0])
        images.append(data_sample[1])
        target.append(data_sample[2])

    indices = torch.tensor(indices, dtype=torch.long)
    images = torch.stack(images, dim=0)
    target = torch.tensor(target, dtype=torch.long)

    images, target = mix_transforms(images, target)

    outputs = (indices, images, target)
    return outputs


def classification_onehot_collate_fn(original_batch, num_classes):
    indices = []
    images = []
    target = []
    for data_sample in original_batch:
        indices.append(data_sample[0])
        images.append(data_sample[1])
        target.append(data_sample[2])

    indices = torch.tensor(indices, dtype=torch.long)
    images = torch.stack(images, dim=0)
    target = torch.tensor(target, dtype=torch.long)
    if -1 not in target:
        target = F.one_hot(target, num_classes=num_classes).to(dtype=images.dtype)

    outputs = (indices, images, target)
    return outputs


def detection_collate_fn(original_batch):
    indices = []
    pixel_values = []
    bbox = []
    label = []
    org_shape = []
    for data_sample in original_batch:
        if 'indices' in data_sample:
            indices.append(data_sample['indices'])
        if 'pixel_values' in data_sample:
            pixel_values.append(data_sample['pixel_values'])
        if 'bbox' in data_sample:
            bbox.append(data_sample['bbox'])
        if 'label' in data_sample:
            label.append(data_sample['label'])
        if 'org_shape' in data_sample:
            org_shape.append(data_sample['org_shape'])
    outputs = {}
    if len(indices) != 0:
        indices = torch.tensor(indices, dtype=torch.long)
        outputs.update({'indices': indices})
    if len(pixel_values) != 0:
        pixel_values = torch.stack(pixel_values, dim=0)
        outputs.update({'pixel_values': pixel_values})
    if len(bbox) != 0:
        outputs.update({'bbox': bbox})
    if len(label) != 0:
        outputs.update({'label': label})
    if len(org_shape) != 0:
        outputs.update({'org_shape': org_shape})

    return outputs



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Any, Dict

import torch

from .registry import PHASE_LIST, TASK_METRIC


class MetricFactory:
    def __init__(self, task, conf_model, **kwargs) -> None:
        self.task = task
        self.conf_model = conf_model

        assert self.task in TASK_METRIC
        self.metric_cls = TASK_METRIC[self.task]

        # TODO: This code assumes there is only one loss module. Fix here later.
        if hasattr(conf_model.losses[0], 'ignore_index'):
            kwargs['ignore_index'] = conf_model.losses[0].ignore_index
        self.metrics = {phase: self.metric_cls(**kwargs) for phase in PHASE_LIST}

    def reset_values(self):
        for phase in PHASE_LIST:
            [meter.reset() for _, meter in self.metrics[phase].metric_meter.items()]

    def update(self, pred: torch.Tensor, target: torch.Tensor, phase: str, **kwargs: Any) -> None:
        if len(pred) == 0: # Removed dummy batch has 0 len
            return
        phase = phase.lower()
        self.metrics[phase].calibrate(pred, target)

    def result(self, phase='train'):
        return {metric_name: meter.avg for metric_name, meter in self.metrics[phase].metric_meter.items()}

    @property
    def metric_names(self):
        return self.metrics[list(self.metrics.keys())[0]].metric_names

    @property
    def primary_metric(self):
        return self.metrics[list(self.metrics.keys())[0]].primary_metric


def build_metrics(task: str, conf_model, **kwargs) -> MetricFactory:
    metric_handler = MetricFactory(task, conf_model, **kwargs)
    return metric_handler
